
#include <SPI.h>
#include <mcp2515.h>
#include <SoftwareSerial.h>
#include "CRC16.h"

const byte SLAVE = 1;
const byte numBytes = 16;
byte receivedBytes[numBytes];
byte responseBytes[64];
byte silence, ndx;
bool done;
byte cmd;
word CRC, startAddr, dataLen;
int registers[150];

struct can_frame canMsg;
struct can_frame canMsg1;
MCP2515 mcp2515(10);

#define TYPE registers[0x0001]
#define VER registers[0x0002]
#define FWH registers[0x0003]
#define FWL registers[0x0004]
#define HWH registers[0x0005]
#define HWL registers[0x0006]
#define NUMCELLS registers[0x0010]
#define CELL1V registers[0x0011]
#define CELL2V registers[0x0012]
#define CELL3V registers[0x0013]
#define CELL4V registers[0x0014]
#define CELL5V registers[0x0015]
#define CELL6V registers[0x0016]
#define CELL7V registers[0x0017]
#define CELL8V registers[0x0018]
#define CELL9V registers[0x0019]
#define CELL10V registers[0x001A]
#define CELL11V registers[0x001B]
#define CELL12V registers[0x001C]
#define CELL13V registers[0x001D]
#define CELL14V registers[0x001E]
#define CELL15V registers[0x001F]
#define CELL16V registers[0x0020]
#define CELL17V registers[0x0021]
#define CELL18V registers[0x0022]
#define CELL19V registers[0x0023]
#define CELL20V registers[0x0024]
#define NUMTEMPS registers[0x0025]
#define TEMP1 registers[0x0026]
#define TEMP2 registers[0x0027]
#define TEMP3 registers[0x0028]
#define TEMP4 registers[0x0029]
#define TEMP5 registers[0x002A]
#define TEMP6 registers[0x002B]
#define TEMP7 registers[0x002C]
#define TEMP8 registers[0x002D]
#define TEMP9 registers[0x002E]
#define TEMP10 registers[0x002F]
#define CHGCUR registers[0x0030]
#define DSCHCUR registers[0x0031]
#define BANKV registers[0x0032]
#define SOC registers[0x0033]
#define BANKCAPH registers[0x0034]
#define BANKCAPL registers[0x0035]
#define NUMCELL registers[0x0040]
#define C1_2VS registers[0x0041]
#define C3_4VS registers[0x0042]
#define C5_6VS registers[0x0043]
#define C7_8VS registers[0x0044]
#define C9_10VS registers[0x0045]
#define C11_12VS registers[0x0046]
#define C13_14VS registers[0x0047]
#define C15_16VS registers[0x0048]
#define C17_18VS registers[0x0049]
#define C19_20VS registers[0x004A]
#define NUMTEMP registers[0x0050]
#define T1_2S registers[0x0051]
#define T3_4S registers[0x0052]
#define T5_6S registers[0x0053]
#define T7_8S registers[0x0054]
#define T9_10S registers[0x0055]
#define BANKCHGVS registers[0x0060]
#define BANKDSCHVS registers[0x0061]
#define CELLCHGVS registers[0x0062]
#define CELLDSCHVS registers[0x0063]
#define BANKCHGCS registers[0x0064]
#define BANKDSCHCS registers[0x0065]
#define BANKCHGTS registers[0x0066]
#define BANKDSCHTS registers[0x0067]
#define CELLCHGTS registers[0x0068]
#define CELLDSCHTS registers[0x0069]
#define CHGVLIM registers[0x0070]
#define DSCHVLIM registers[0x0071]
#define CHGCLIM registers[0x0072]
#define DSCHCLIM registers[0x0073]
#define CHGDSCHSTS registers[0x0074]
// CHGEN:DSCHEN:CHGNOW:CHGNOW2:FULLCHG:0:0:0

SoftwareSerial EG4(4, 3); // RX, TX
CRC16 crc(CRC16_MODBUS_POLYNOME,
          CRC16_MODBUS_INITIAL,
          CRC16_MODBUS_XOR_OUT,
          CRC16_MODBUS_REV_IN,
          CRC16_MODBUS_REV_OUT);

void setup() {
  TYPE = 0;
  VER = 0;
  FWH = 0;
  FWL = 0;
  HWH = 0;
  HWL = 0;
  NUMCELLS = 16;
  CELL1V = 3000;
  CELL2V = 3000;
  CELL3V = 3000;
  CELL4V = 3000;
  CELL5V = 3000;
  CELL6V = 3000;
  CELL7V = 3000;
  CELL8V = 3000;
  CELL9V = 3000;
  CELL10V = 3000;
  CELL11V = 3000;
  CELL12V = 3000;
  CELL13V = 3000;
  CELL14V = 3000;
  CELL15V = 3000;
  CELL16V = 3000;
  CELL17V = 0;
  CELL18V = 0;
  CELL19V = 0;
  CELL20V = 0;
  NUMTEMPS = 3;
  TEMP1 = 3000;
  TEMP2 = 3000;
  TEMP3 = 3000;
  TEMP4 = 0;
  TEMP5 = 0;
  TEMP6 = 0;
  TEMP7 = 0;
  TEMP8 = 0;
  TEMP9 = 0;
  TEMP10 = 0;
  CHGCUR = 20;
  DSCHCUR = 100;
  BANKV = 480;
  SOC = 50;
  BANKCAPH = 0;
  BANKCAPL = 50000;
  NUMCELL = 16;
  C1_2VS = 0;
  C3_4VS = 0;
  C5_6VS = 0;
  C7_8VS = 0;
  C9_10VS = 0;
  C11_12VS = 0;
  C13_14VS = 0;
  C15_16VS = 0;
  C17_18VS = 0;
  C19_20VS = 0;
  NUMTEMP = 3;
  T1_2S = 0;
  T3_4S = 0;
  T5_6S = 0;
  T7_8S = 0;
  T9_10S = 0;
  BANKCHGVS = 0;
  BANKDSCHVS = 0;
  CELLCHGVS = 0;
  CELLDSCHVS = 0;
  BANKCHGCS = 0;
  BANKDSCHCS = 0;
  BANKCHGTS = 0;
  BANKDSCHTS = 0;
  CELLCHGTS = 0;
  CELLDSCHTS = 0;
  CHGVLIM = 560;
  DSCHVLIM = 480;
  CHGCLIM = 20;
  DSCHCLIM = 500;
  CHGDSCHSTS = 0b11000000; // CHGEN:DSCHEN:CHGNOW:CHGNOW2:FULLCHG:0:0:0

  canMsg1.can_id  = 0x305; // heartbeat to BMS
  canMsg1.can_dlc = 8;
  canMsg1.data[0] = 0x00;
  canMsg1.data[1] = 0x00;
  canMsg1.data[2] = 0x00;
  canMsg1.data[3] = 0x00;
  canMsg1.data[4] = 0x00;
  canMsg1.data[5] = 0x00;
  canMsg1.data[6] = 0x00;
  canMsg1.data[7] = 0x00;

// Open serial communications and wait for port to open:
  Serial.begin(115200);
  while (!Serial) {
    ; // wait for serial port to connect. Needed for native USB port only
  }

  mcp2515.reset();
  mcp2515.setBitrate(CAN_500KBPS, MCP_8MHZ); // 500kbps works for bms
  mcp2515.setNormalMode();

// set the data rate for the SoftwareSerial port
  EG4.begin(9600);
}

void loop() {

  if (mcp2515.readMessage(&canMsg) == MCP2515::ERROR_OK) {
    Serial.print(canMsg.can_id, HEX); // print ID
    Serial.print(F(" ")); Serial.print(canMsg.can_dlc, HEX); // print DLC
    Serial.print(F(" "));
    for (int i = 0; i<canMsg.can_dlc; i++)  {  // print the data
      Serial.print(canMsg.data[i],HEX); Serial.print(F(" "));
    }
    switch (canMsg.can_id) {
      case 0x359:
        Serial.print(F("Alarm Bits: 8 bytes expected, ")); Serial.print(canMsg.can_dlc, HEX); Serial.println(F(" bytes received"));
//        Serial.print(F("Byte 0, Bit 0: ")); Serial.println(bitRead(canMsg.data[0],0));
        Serial.print(F("Byte 0, Bit 1, cell over volt: ")); Serial.print(bitRead(canMsg.data[0],1));
        Serial.print(F("  Byte 0, Bit 2, cell under volt: ")); Serial.println(bitRead(canMsg.data[0],2));
        Serial.print(F("Byte 0, Bit 3, cell over temp: ")); Serial.print(bitRead(canMsg.data[0],3));
        Serial.print(F("  Byte 0, Bit 4, cell under temp: ")); Serial.println(bitRead(canMsg.data[0],4));
//        Serial.print(F("Byte 0, Bit 5: ")); Serial.println(bitRead(canMsg.data[0],5));
//        Serial.print(F("Byte 0, Bit 6: ")); Serial.println(bitRead(canMsg.data[0],6));
        Serial.print(F("Byte 0, Bit 7, disch overcurrent: ")); Serial.print(bitRead(canMsg.data[0],7));
        Serial.print(F("  Byte 1, Bit 0, chg overcurrent: ")); Serial.println(bitRead(canMsg.data[1],0));
//        Serial.print(F("Byte 1, Bit 1: ")); Serial.println(bitRead(canMsg.data[1],1));
//        Serial.print(F("Byte 1, Bit 2: ")); Serial.println(bitRead(canMsg.data[1],2));
        Serial.print(F("Byte 1, Bit 3, Sys err: ")); Serial.println(bitRead(canMsg.data[1],3));
//        Serial.print(F("Byte 1, Bit 4: ")); Serial.println(bitRead(canMsg.data[1],4));
//        Serial.print(F("Byte 1, Bit 5: ")); Serial.println(bitRead(canMsg.data[1],5));
//        Serial.print(F("Byte 1, Bit 6: ")); Serial.println(bitRead(canMsg.data[1],6));
//        Serial.print(F("Byte 1, Bit 7: ")); Serial.println(bitRead(canMsg.data[1],7));
//        Serial.print(F("Byte 2, Bit 0: ")); Serial.println(bitRead(canMsg.data[2],0));
        Serial.print(F("Byte 2, Bit 1, cell high volt: ")); Serial.print(bitRead(canMsg.data[2],1));
        Serial.print(F("  Byte 2, Bit 2, cell low volt: ")); Serial.println(bitRead(canMsg.data[2],2));
        Serial.print(F("Byte 2, Bit 3, cell high temp: ")); Serial.print(bitRead(canMsg.data[2],3));
        Serial.print(F("  Byte 2, Bit 4, cell low temp: ")); Serial.println(bitRead(canMsg.data[2],4));
//        Serial.print(F("Byte 2, Bit 5: ")); Serial.println(bitRead(canMsg.data[2],5));
//        Serial.print(F("Byte 2, Bit 6: ")); Serial.println(bitRead(canMsg.data[2],6));
        Serial.print(F("Byte 2, Bit 7, disch high current: ")); Serial.print(bitRead(canMsg.data[2],7));
        Serial.print(F("  Byte 3, Bit 0, chg high current: ")); Serial.println(bitRead(canMsg.data[3],0));
//        Serial.print(F("Byte 3, Bit 1: ")); Serial.println(bitRead(canMsg.data[3],1));
//        Serial.print(F("Byte 3, Bit 2: ")); Serial.println(bitRead(canMsg.data[3],2));
        Serial.print(F("Byte 3, Bit 3, internal comm fail: ")); Serial.println(bitRead(canMsg.data[3],3));
//        Serial.print(F("Byte 3, Bit 4: ")); Serial.println(bitRead(canMsg.data[3],4));
//        Serial.print(F("Byte 3, Bit 5: ")); Serial.println(bitRead(canMsg.data[3],5));
//        Serial.print(F("Byte 3, Bit 6: ")); Serial.println(bitRead(canMsg.data[3],6));
//        Serial.print(F("Byte 3, Bit 7: ")); Serial.println(bitRead(canMsg.data[3],7));
        BANKCAPL = (canMsg.data[4]*74000)&0xFF; BANKCAPH = ((canMsg.data[4]*74000)>>8)&0xFF;
        Serial.print(F("Byte 4, # of Modules: ")); Serial.print(canMsg.data[4]);
        Serial.print(F("  Byte 5, 'P': ")); Serial.print((char)canMsg.data[5]);
        Serial.print(F("  Byte 6, 'N': ")); Serial.println((char)canMsg.data[6]);
//        Serial.print(F("Byte 7: ")); Serial.println(canMsg.data[7]);
        
      break;
      case 0x351:
        Serial.print(F("Charge Voltage & Current Limits: 8 bytes expected, ")); Serial.print(canMsg.can_dlc, HEX); Serial.println(F(" bytes received"));
        if (CHGVLIM != word(canMsg.data[1],canMsg.data[0])) {
          Serial.print(F("CHGVLIM updated from ")); Serial.print(CHGVLIM);
          CHGVLIM = word(canMsg.data[1],canMsg.data[0]);
          Serial.print(F(" to ")); Serial.println(CHGVLIM);
        }
//        Serial.print(F("Byte 0&1, batt chg voltage: ")); Serial.println(((int)word(canMsg.data[1],canMsg.data[0]))/10.0);
        if (CHGCLIM != word(canMsg.data[3],canMsg.data[2])) {
          Serial.print(F("CHGCLIM updated from ")); Serial.print(CHGCLIM);
          CHGCLIM = word(canMsg.data[3],canMsg.data[2]);
          Serial.print(F(" to ")); Serial.println(CHGCLIM);
        }
//        Serial.print(F("Byte 2&3, chg curr limit: ")); Serial.println(((int)word(canMsg.data[3],canMsg.data[2]))/10.0);
        if (DSCHCLIM != word(canMsg.data[5],canMsg.data[4])) {
          Serial.print(F("DSCHCLIM updated from ")); Serial.print(DSCHCLIM);
          DSCHCLIM = word(canMsg.data[5],canMsg.data[4]);
          Serial.print(F(" to ")); Serial.println(DSCHCLIM);
        }
//        Serial.print(F("Byte 4&5, disch curr limit: ")); Serial.println(((int)word(canMsg.data[5],canMsg.data[4]))/10.0);
        if (DSCHVLIM != word(canMsg.data[7],canMsg.data[6])) {
          Serial.print(F("DSCHVLIM updated from ")); Serial.print(DSCHVLIM);
          DSCHVLIM = word(canMsg.data[7],canMsg.data[6]);
          Serial.print(F(" to ")); Serial.println(DSCHVLIM);
        }
//        Serial.print(F("Byte 6&7, batt disch voltage: ")); Serial.println(((int)word(canMsg.data[7],canMsg.data[6]))/10.0);
      break;
      case 0x355:
        Serial.print(F("SOC/SOH%: 8 bytes expected, ")); Serial.print(canMsg.can_dlc, HEX); Serial.println(F(" bytes received"));
        if (SOC != word(canMsg.data[1],canMsg.data[0])) {
          Serial.print(F("SOC updated from ")); Serial.print(SOC);
          SOC = word(canMsg.data[1],canMsg.data[0]);
          Serial.print(F(" to ")); Serial.println(SOC);
        }
//        Serial.print(F("Byte 0&1, SOC: ")); Serial.println((int)word(canMsg.data[1],canMsg.data[0]));
//        Serial.print(F("Byte 2&3, SOH: ")); Serial.println((int)word(canMsg.data[3],canMsg.data[2]));
//        Serial.print(F("Byte 4: ")); Serial.println(canMsg.data[4]);
//        Serial.print(F("Byte 5: ")); Serial.println(canMsg.data[5]);
//        Serial.print(F("Byte 6: ")); Serial.println(canMsg.data[6]);
//        Serial.print(F("Byte 7: ")); Serial.println(canMsg.data[7]);
      break;
      case 0x356:
        Serial.print(F("Module Stats: 8 bytes expected, ")); Serial.print(canMsg.can_dlc, HEX); Serial.println(F(" bytes received"));
        Serial.print(F("Byte 0&1, bank voltage: ")); Serial.print((int)word(canMsg.data[1],canMsg.data[0])/100.0);
        Serial.print(F("  Byte 2&3, bank current: ")); Serial.print((int)word(canMsg.data[3],canMsg.data[2])/10.0);
        Serial.print(F("  Byte 4&5, avg cell temp: ")); Serial.println((int)word(canMsg.data[5],canMsg.data[4])/10.0);
//        Serial.print(F("Byte 6: ")); Serial.println(canMsg.data[6]);
//        Serial.print(F("Byte 7: ")); Serial.println(canMsg.data[7]);
      break;
      case 0x35C:
        Serial.print(F("Request Flags: 2 bytes expected, ")); Serial.print(canMsg.can_dlc, HEX); Serial.println(F(" bytes received"));
        if (CHGDSCHSTS != canMsg.data[0]) {
          Serial.print(F("CHGDSCHSTS updated from ")); Serial.print(CHGDSCHSTS);
          CHGDSCHSTS = canMsg.data[0];
          Serial.print(F(" to ")); Serial.println(CHGDSCHSTS);
        }
//        Serial.print(F("Byte 0, Bit 0: ")); Serial.println(bitRead(canMsg.data[0],0));
//        Serial.print(F("Byte 0, Bit 1: ")); Serial.println(bitRead(canMsg.data[0],1));
//        Serial.print(F("Byte 0, Bit 2: ")); Serial.println(bitRead(canMsg.data[0],2));
/*        Serial.print(F("Byte 0, Bit 3, full chg req: ")); Serial.print(bitRead(canMsg.data[0],3));
        Serial.print(F("  Byte 0, Bit 4, force chg req II: ")); Serial.print(bitRead(canMsg.data[0],4));
        Serial.print(F("  Byte 0, Bit 5, force chg req I: ")); Serial.println(bitRead(canMsg.data[0],5));
        Serial.print(F("Byte 0, Bit 6, disch enable: ")); Serial.print(bitRead(canMsg.data[0],6));
        Serial.print(F("  Byte 0, Bit 7, chg enable: ")); Serial.println(bitRead(canMsg.data[0],7));*/
//        Serial.print(F("Byte 1, Bit 0: ")); Serial.println(bitRead(canMsg.data[1],0));
//        Serial.print(F("Byte 1, Bit 1: ")); Serial.println(bitRead(canMsg.data[1],1));
//        Serial.print(F("Byte 1, Bit 2: ")); Serial.println(bitRead(canMsg.data[1],2));
//        Serial.print(F("Byte 1, Bit 3: ")); Serial.println(bitRead(canMsg.data[1],3));
//        Serial.print(F("Byte 1, Bit 4: ")); Serial.println(bitRead(canMsg.data[1],4));
//        Serial.print(F("Byte 1, Bit 5: ")); Serial.println(bitRead(canMsg.data[1],5));
//        Serial.print(F("Byte 1, Bit 6: ")); Serial.println(bitRead(canMsg.data[1],6));
//        Serial.print(F("Byte 1, Bit 7: ")); Serial.println(bitRead(canMsg.data[1],7));
      break;
      case 0x35E:
        Serial.print(F("Manuf: 8 bytes expected, ")); Serial.print(canMsg.can_dlc, HEX); Serial.println(F(" bytes received"));
        Serial.print(F("Bytes: ")); Serial.print((char)canMsg.data[0]); Serial.print((char)canMsg.data[1]);
        Serial.print((char)canMsg.data[2]); Serial.print((char)canMsg.data[3]);
        Serial.print((char)canMsg.data[4]); Serial.print((char)canMsg.data[5]);
        Serial.print((char)canMsg.data[6]); Serial.println((char)canMsg.data[7]);
      break;
      case 0x305:
        Serial.print(F("Inverter Heartbeat: 8 bytes expected, ")); Serial.print(canMsg.can_dlc, HEX); Serial.println(F(" bytes received"));
        Serial.print(F("Byte 0: ")); Serial.println(canMsg.data[0]); Serial.print(F("Byte 1: ")); Serial.println(canMsg.data[1]);
        Serial.print(F("Byte 2: ")); Serial.println(canMsg.data[2]); Serial.print(F("Byte 3: ")); Serial.println(canMsg.data[3]);
        Serial.print(F("Byte 4: ")); Serial.println(canMsg.data[4]); Serial.print(F("Byte 5: ")); Serial.println(canMsg.data[5]);
        Serial.print(F("Byte 6: ")); Serial.println(canMsg.data[6]); Serial.print(F("Byte 7: ")); Serial.println(canMsg.data[7]);
      break;
      default:
        Serial.println(F("CAN ID not recognized!"));
      break;
      
    }
    Serial.println();      
  }
  
  while (EG4.available()) {
    silence = 0; done = false;
    receivedBytes[ndx] = EG4.read();
    Serial.print(receivedBytes[ndx],HEX); Serial.print(F(" "));
    ndx += 1;
  }
  if (!done) {
    silence += 1;
    if (silence == 100) { // it's been quiet for a while, so EG4 must be done transmitting
      Serial.print('\t'); // tab
      done = true; // EG4 done transmitting command
      processData();
      crc.restart();
      ndx = 0;
    }
  }
}

void processData() {
  cmd = receivedBytes[1];
  startAddr = word(receivedBytes[2],receivedBytes[3]);
  Serial.print(F("S:")); Serial.print(startAddr,HEX);
  dataLen = word(receivedBytes[4],receivedBytes[5]);
  Serial.print(F(" L:")); Serial.print(dataLen,HEX);
  for (int i=0; i < (ndx-2); i++) crc.add(receivedBytes[i]); // include all but last two received bytes in crc calculation
  CRC = word(receivedBytes[ndx-1],receivedBytes[ndx-2]);
  ndx = 0; // begin building response
  responseBytes[ndx] = SLAVE; ndx++;
  if (CRC == crc.calc()) { // transmitted crc matches calculated crc
    Serial.print(F(" good crc! "));
    if (receivedBytes[0] && receivedBytes[0]<17) { // 0 < slade id < 17
      Serial.print(F("good slave!")); Serial.print('\t');
      responseBytes[ndx] = cmd; ndx++;
      switch (cmd) {
        case 3: // read
          responseBytes[ndx] = highByte(dataLen); crc.add(responseBytes[ndx]); ndx += 1;
          responseBytes[ndx] = lowByte(dataLen); crc.add(responseBytes[ndx]); ndx += 1;
          for (int addr = startAddr; addr < (startAddr + dataLen); addr++) {
            responseBytes[ndx] = highByte(registers[addr]); crc.add(responseBytes[ndx]); ndx += 1; // data
            responseBytes[ndx] = lowByte(registers[addr]); crc.add(responseBytes[ndx]); ndx += 1; // data
          }
        break;
//        case 10: // write
//          Serial.println(F("write cmd.")); // not currently implemented, may break code if write cmd received. ignore for now.
//          break;
        default:
          Serial.println(F("unrecognized/unused cmd."));
          cmd += 128; responseBytes[ndx] = cmd; ndx++; responseBytes[ndx] = 2; // cmd error
        break;
      }
    }
    else {
      Serial.println(F("BAD slave id!"));
      cmd += 128; responseBytes[ndx] = cmd; ndx++; responseBytes[ndx] = 1; ndx++; } // slave error
  }
  else {
    Serial.println(F("BAD crc!"));
    cmd += 128; responseBytes[ndx] = cmd; ndx++; responseBytes[ndx] = 3; ndx++;} // crc error
  crc.restart(); for (int i=0; i < (ndx); i++) crc.add(responseBytes[i]); // include all response bytes in crc calculation
  responseBytes[ndx] = lowByte(crc.calc()); ndx += 1; responseBytes[ndx] = highByte(crc.calc()); ndx += 1;
  if (receivedBytes[0] == SLAVE) { // they're talking to us!
    Serial.print(F("response:"));
    for (byte i=0; i<ndx; i++) {
      Serial.print(F(" ")); Serial.print(responseBytes[i],HEX);
      EG4.write(responseBytes[i]); // transmit response
    }
    Serial.println(); EG4.write('\0'); // if needed
    mcp2515.sendMessage(&canMsg1); // send heartbeat to BMS to confirm EG4 traffic
  }
  else {Serial.println(F("msg not meant for us"));
  }
}

uint8_t toHex(char hi, char lo) {
  uint8_t b;
  hi = toupper(hi);
  if( isxdigit(hi) ) {
    if( hi > '9' ) hi -= 7;      // software offset for A-F
    hi -= 0x30;                  // subtract ASCII offset
    b = hi<<4;
    lo = toupper(lo);
    if( isxdigit(lo) ) {
      if( lo > '9' ) lo -= 7;  // software offset for A-F
      lo -= 0x30;              // subtract ASCII offset
      b = b + lo;
      return b;
    } // else error
  }  // else error
  return 0;
}
